@@ -77,9 +77,9 @@ test_that("bootstrap function can detect PyH errors correctly.",{
   skip_if(T)
   ripser = import_ripser()
   D <- TDA::circleUnif(n = 50,r = 1)
-  expect_error(bootstrap_persistence_thresholds(X = D,FUN = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = NULL),"NULL")
-  expect_error(bootstrap_persistence_thresholds(X = D,FUN = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = 2),"logical")
-  expect_error(bootstrap_persistence_thresholds(X = D,FUN = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = NA),"NA")
+  expect_error(bootstrap_persistence_thresholds(X = D,FUN_diag = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = NULL),"NULL")
+  expect_error(bootstrap_persistence_thresholds(X = D,FUN_diag = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = 2),"logical")
+  expect_error(bootstrap_persistence_thresholds(X = D,FUN_boot = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = NA),"NA")
   
 })
 
@@ -90,7 +90,7 @@ test_that("PyH functionality works in bootstrap function.",{
   D <- TDA::circleUnif(n = 50,r = 1)
   
   # PyH with multiple thresholds
-  bs <- bootstrap_persistence_thresholds(X = D,FUN = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = F)
+  bs <- bootstrap_persistence_thresholds(X = D,FUN_diag = "PyH",FUN_boot = "PyH",maxdim = 1,thresh = 2,calculate_representatives = T,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3,return_subsetted = T,ignore_infinite_cluster = F)
   expect_length(bs$representatives[[2]],length(which(bs$diag$dimension == 1)))
   expect_length(bs$thresholds,2)
   expect_gt(bs$thresholds[[1]],0)
@@ -104,9 +104,9 @@ test_that("PyH functionality works in bootstrap function.",{
   expect_true(min(bs$subsetted_diag[which(bs$subsetted_diag$dimension == 1),]$death - bs$subsetted_diag[which(bs$subsetted_diag$dimension == 1),]$birth) > bs$thresholds[[2]])
   
   # check on circle
-  bs <- bootstrap_persistence_thresholds(X = D,FUN = "PyH",maxdim = 1,thresh = 2,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3)
+  bs <- bootstrap_persistence_thresholds(X = D,FUN_diag = "PyH",maxdim = 1,thresh = 2,return_diag = T,ripser = ripser,num_workers = 2,num_samples = 3)
   expect_lte(length(bs$subsetted_diag$dimension),1)
-  bs <- bootstrap_persistence_thresholds(X = D,FUN = "PyH",maxdim = 1,thresh = 2,return_diag = T,ripser = ripser,ignore_infinite_cluster = F,num_workers = 2,num_samples = 3)
+  bs <- bootstrap_persistence_thresholds(X = D,FUN_diag = "PyH",maxdim = 1,thresh = 2,return_diag = T,ripser = ripser,ignore_infinite_cluster = F,num_workers = 2,num_samples = 3)
   expect_lte(length(bs$subsetted_diag$dimension),2)
 
 })
